import json
import os


# TODO: this function is copy-pasted from login.py;
#  there is something weird going one with imports in __init__.py and I'm unable to import apikey now
def get_api_key(config_dir_path, config_file_name):
    paperspace_dir = os.path.expanduser(config_dir_path)
    config_path = os.path.join(paperspace_dir, config_file_name)
    if os.path.exists(config_path):
        config_data = json.load(open(config_path))
        if config_data and 'apiKey' in config_data:
            return config_data['apiKey']
    return ''


_DEFAULT_WEB_URL = "https://console.paperspace.com"
_DEFAULT_CONFIG_HOST = "https://api.paperspace.io"
_DEFAULT_CONFIG_LOG_HOST = "https://logs.paperspace.io"
_DEFAULT_CONFIG_EXPERIMENTS_HOST = "https://services.paperspace.io/experiments/v1/"
_DEFAULT_CONFIG_EXPERIMENTS_HOST_V2 = "https://services.paperspace.io/experiments/v2/"
_DEFAULT_CONFIG_SERVICE_HOST = "https://services.paperspace.io"
_DEFAULT_CONFIG_DIR_PATH = "~/.paperspace"
_DEFAULT_CONFIG_FILE_NAME = os.path.expanduser("config.json")
_DEFAULT_HELP_HEADERS_COLOR = "yellow"
_DEFAULT_HELP_OPTIONS_COLOR = "green"
_DEFAULT_USE_CONSOLE_COLORS = True


def get_help_colors_dict(use_colors, help_headers_color, help_options_color):
    if not use_colors:
        return {}

    d = {
        "help_headers_color": help_headers_color,
        "help_options_color": help_options_color,
    }
    return d


class config(object):
    DEBUG = os.environ.get("PAPERSPACE_CLI_DEBUG") in ("true", "1")

    WEB_URL = os.environ.get("PAPERSPACE_WEB_URL", _DEFAULT_WEB_URL)
    CONFIG_HOST = os.environ.get("PAPERSPACE_CONFIG_HOST", _DEFAULT_CONFIG_HOST)
    CONFIG_LOG_HOST = os.environ.get("PAPERSPACE_CONFIG_LOG_HOST", _DEFAULT_CONFIG_LOG_HOST)
    CONFIG_EXPERIMENTS_HOST = os.environ.get("PAPERSPACE_CONFIG_EXPERIMENTS_HOST", _DEFAULT_CONFIG_EXPERIMENTS_HOST)
    CONFIG_EXPERIMENTS_HOST_V2 = os.environ.get("PAPERSPACE_CONFIG_EXPERIMENTS_HOST_V2",
                                                _DEFAULT_CONFIG_EXPERIMENTS_HOST_V2)
    CONFIG_SERVICE_HOST = os.environ.get("PAPERSPACE_CONFIG_SERVICE_HOST", _DEFAULT_CONFIG_SERVICE_HOST)
    CONFIG_DIR_PATH = os.path.expanduser(os.environ.get("PAPERSPACE_CONFIG_PATH", _DEFAULT_CONFIG_DIR_PATH))
    CONFIG_FILE_NAME = os.environ.get("PAPERSPACE_CONFIG_FILE_NAME", _DEFAULT_CONFIG_FILE_NAME)
    PAPERSPACE_API_KEY = os.environ.get("PAPERSPACE_API_KEY", get_api_key(CONFIG_DIR_PATH, CONFIG_FILE_NAME))

    HELP_HEADERS_COLOR = os.environ.get("PAPERSPACE_HELP_HEADERS_COLOR", _DEFAULT_HELP_HEADERS_COLOR)
    HELP_OPTIONS_COLOR = os.environ.get("PAPERSPACE_HELP_OPTIONS_COLOR", _DEFAULT_HELP_OPTIONS_COLOR)
    USE_CONSOLE_COLORS = os.environ.get("PAPERSPACE_USE_CONSOLE_COLORS",
                                        _DEFAULT_USE_CONSOLE_COLORS) in (True, "true", "1")
    HELP_COLORS_DICT = get_help_colors_dict(USE_CONSOLE_COLORS, HELP_HEADERS_COLOR, HELP_OPTIONS_COLOR)



import attr


@attr.s
class Model(object):
    """
    Model class

    :param str id:
    :param str name:
    :param str project_id:
    :param str experiment_id:
    :param str cluster_id:
    :param list tags:
    :param str model_type:
    :param str url:
    :param str model_path:
    :param str deployment_state:
    :param str summary:
    :param str detail:
    :param str dataset_ref:
    """
    id = attr.ib(type=str, default=None)
    name = attr.ib(type=str, default=None)
    project_id = attr.ib(type=str, default=None)
    experiment_id = attr.ib(type=str, default=None)
    cluster_id = attr.ib(type=str, default=None)
    tags = attr.ib(type=list, factory=list)
    model_type = attr.ib(type=str, default=None)
    url = attr.ib(type=str, default=None)
    model_path = attr.ib(type=str, default=None)
    deployment_state = attr.ib(type=str, default=None)
    summary = attr.ib(type=dict, default=None)
    detail = attr.ib(type=dict, default=None)
    notes = attr.ib(type=str, default=None)
    dataset_ref = attr.ib(type=str, default=None)


@attr.s
class ModelFile(object):
    """
    Model file

    :param str file: file path
    :param str url: Url with AWS key
    :param int size: File size in bytes
    """
    file = attr.ib(type=str, default=None)
    url = attr.ib(type=str, default=None)
    size = attr.ib(type=int, default=None)



import functools

import click

from gradient.cli import common
from gradient.cli.common import ClickGroup, validate_comma_split_option
from gradient.cli.experiments import common_experiments_create_options, get_workspace_handler, experiments_group
from gradient.commands import hyperparameters as hyperparameters_commands
from gradient.commands.hyperparameters import HyperparameterAddTagsCommand, HyperparameterRemoveTagsCommand


@experiments_group.group("hyperparameters", help="Manage hyperparameters", cls=ClickGroup)
def hyperparameters_group():
    pass


@hyperparameters_group.group("tags", help="Manage hyperparameter tags", cls=ClickGroup)
def hyperparameters_tags():
    pass


def common_hyperparameter_create_options(f):
    options = [
        click.option(
            "--tuningCommand",
            "tuning_command",
            required=True,
            help="Tuning command",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerContainer",
            "worker_container",
            required=True,
            help="Worker container",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerContainerUser",
            "worker_container_user",
            required=False,
            help="Worker container user",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerMachineType",
            "worker_machine_type",
            required=True,
            help="Worker machine type",
            cls=common.GradientOption,
        ),
        click.option(
            "--hyperparameterServerMachineType",
            "hyperparameter_server_machine_type",
            required=False,
            help="Hyperparameter Server machine type",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerCommand",
            "worker_command",
            required=True,
            help="Worker command",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerCount",
            "worker_count",
            required=True,
            type=int,
            help="Worker count",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerUseDockerfile",
            "use_dockerfile",
            type=bool,
            is_flag=True,
            default=False,
            help="Flag: use dockerfile",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerDockerfilePath",
            "dockerfile_path",
            help="Path to ",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerRegistryUsername",
            "worker_registry_username",
            help="Worker registry username",
            cls=common.GradientOption,
        ),
        click.option(
            "--workerRegistryPassword",
            "worker_registry_password",
            help="Worker registry password",
            cls=common.GradientOption,
        ),
        click.option(
            "--hyperparameterServerRegistryUsername",
            "hyperparameter_server_registry_username",
            help="Hyperparameter server registry username",
            cls=common.GradientOption,
        ),
        click.option(
            "--hyperparameterServerRegistryPassword",
            "hyperparameter_server_registry_password",
            help="Hyperparameter server registry password",
            cls=common.GradientOption,
        ),
        click.option(
            "--hyperparameterServerContainer",
            "hyperparameter_server_container",
            help="Hyperparameter server container",
            cls=common.GradientOption,
        ),
        click.option(
            "--hyperparameterServerContainerUser",
            "hyperparameter_server_container_user",
            help="Hyperparameter server container user",
            cls=common.GradientOption,
        ),
    ]
    return functools.reduce(lambda x, opt: opt(x), reversed(options), f)


@hyperparameters_group.command("create", help="Create hyperparameter")
@common_experiments_create_options
@common_hyperparameter_create_options
@common.api_key_option
@common.options_file
def create_hyperparameter(api_key, options_file, **hyperparameter):
    hyperparameter["tags"] = validate_comma_split_option(hyperparameter.pop("tags_comma"), hyperparameter.pop("tags"))
    common.del_if_value_is_none(hyperparameter, del_all_falsy=True)

    command = hyperparameters_commands.CreateHyperparameterCommand(
        api_key=api_key,
        workspace_handler=get_workspace_handler(api_key),
    )
    command.execute(hyperparameter)


@hyperparameters_group.command("run", help="Create and start hyperparameter tuning job")
@common_experiments_create_options
@common_hyperparameter_create_options
@common.api_key_option
@common.options_file
def create_and_start_hyperparameter(api_key, options_file, **hyperparameter):
    hyperparameter["tags"] = validate_comma_split_option(hyperparameter.pop("tags_comma"), hyperparameter.pop("tags"))
    common.del_if_value_is_none(hyperparameter, del_all_falsy=True)

    command = hyperparameters_commands.CreateAndStartHyperparameterCommand(
        api_key=api_key,
        workspace_handler=get_workspace_handler(api_key),
    )
    command.execute(hyperparameter)


@hyperparameters_group.command("list", help="List hyperparameters")
@common.api_key_option
@common.options_file
def list_hyperparameters(api_key, options_file):
    command = hyperparameters_commands.ListHyperparametersCommand(api_key=api_key)
    command.execute()


@hyperparameters_group.command("details", help="Show details of hyperparameter")
@click.option(
    "--id",
    "id_",
    required=True,
    cls=common.GradientOption,
)
@common.api_key_option
@common.options_file
def get_hyperparameter_details(api_key, id_, options_file):
    command = hyperparameters_commands.HyperparameterDetailsCommand(api_key=api_key)
    command.execute(id_)


@hyperparameters_group.command("start", help="Start hyperparameter tuning")
@click.option(
    "--id",
    "id_",
    required=True,
    cls=common.GradientOption,
)
@common.api_key_option
@common.options_file
def start_hyperparameter_tuning(api_key, options_file, id_):
    command = hyperparameters_commands.HyperparameterStartCommand(api_key=api_key)
    command.execute(id_)


@hyperparameters_tags.command("add", help="Add tags to hyperparameter")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the hyperparameter",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to hyperparameter",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to hyperparameter",
    cls=common.GradientOption
)
@common.api_key_option
@common.options_file
def hyperparameter_add_tag(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = HyperparameterAddTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)


@hyperparameters_tags.command("remove", help="Remove tags from hyperparameter")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the hyperparameter",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to remove from hyperparameter",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want to remove from hyperparameter",
    cls=common.GradientOption
)
@common.api_key_option
@common.options_file
def hyperparameter_remove_tags(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = HyperparameterRemoveTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)



"""
Jobs related client handler logic.

Remember that in code snippets all highlighted lines are required other lines are optional.
"""
from .base_client import BaseClient, TagsSupportMixin
from ..models import Artifact, Job
from ..repositories.jobs import ListJobs, ListJobLogs, ListJobArtifacts, CreateJob, DeleteJob, StopJob, \
    DeleteJobArtifacts, GetJobArtifacts, GetJobMetrics, ListJobMetrics, StreamJobMetrics


class JobsClient(TagsSupportMixin, BaseClient):
    """
    Client to handle job related actions.

    How to create instance of job client:

    .. code-block:: python
        :linenos:
        :emphasize-lines: 4

        from gradient import JobsClient

        job_client = JobClient(
            api_key='your_api_key_here'
        )

    """
    entity = "job"

    def create(
            self,
            machine_type,
            container,
            project_id,
            data=None,
            name=None,
            command=None,
            ports=None,
            is_public=None,
            working_directory=None,
            experiment_id=None,
            env_vars=None,
            use_dockerfile=None,
            is_preemptible=None,
            project=None,
            started_by_user_id=None,
            rel_dockerfile_path=None,
            registry_username=None,
            registry_password=None,
            cluster=None,
            cluster_id=None,
            node_attrs=None,
            workspace_file_name=None,
            registry_target=None,
            registry_target_username=None,
            registry_target_password=None,
            build_only=False,
            tags=None,
            datasets=None,
    ):
        """
        Method to create and start job in paperspace gradient.

        Example create job:

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2,3,4

            job = job_client.create(
                machine_type='K80',
                container='tensorflow/tensorflow:1.13.1-gpu-py3',
                project_id='Som3ProjecTiD',
                data=data,
                name='Example job',
                command='pip install -r requirements.txt && python mnist.py',
                ports='5000:5000',
                env_vars={'CUSTOM_ENV': 'Some value that will be set as system environment'}
            )

        :param str machine_type: Type of machine on which job should run. This field is **required**.

            We recommend to choose one of this:

            .. code-block::

                K80
                P100
                TPU
                GV100
                GV100x8
                G1
                G6
                G12

        :param str container: name of docker container that should be used to run job. This field is **required**.

            Example value: ``tensorflow/tensorflow:1.13.1-gpu-py3``

        :param str project_id: Identify to which project job should be connected. This field is **required**.

        :param None|MultipartEncoderMonitor data: None if there are no data to upload or
            encoded multipart data information with files to upload.
        :param str name: name for job that creator wish to have. If not provided it will be autogenerated.
        :param str command: custom command that should be run against command from docker image
        :param str ports: string with comma `,` separated mapped ports.

            Example value: ``5000:5000,8080:8080``

        :param bool is_public: bool flag to select if job should be available by default None
        :param str working_directory: location of code to run. By default ``/paperspace``
        :param str experiment_id: Id of experiment to which job should be connected. If not provided there will be
            created new experiment for this job.
        :param dict env_vars: key value collection of envs that are used in code
        :param bool use_dockerfile: determines whether to build from Dockerfile (default false).
            Do not include a --container argument when using this flag.
        :param bool is_preemptible: flag if we you want to use spot instance. By default False
        :param str project: name of project that job is linked to.
        :param str started_by_user_id: id of user that started job. By default it take user id from access token
            or api key.
        :param str rel_dockerfile_path: relative location to your dockerfile. Default set to ``./Dockerfile``
        :param str registry_username: username for custom docker registry
        :param str registry_password: password for custom docker registry
        :param str cluster: name of cluster that job should be run on.
        :param str cluster_id: id of cluster that job should be run on. If you use one of recommended machine type
            cluster will be chosen so you do not need to provide it.
        :param dict node_attrs:
        :param str workspace_file_name:
        :param str registry_target: custom docker registry
        :param str registry_target_username: username for custom docker registry
        :param str registry_target_password: password for custom docker registry
        :param bool build_only: determines whether to only build and not run image
        :param list[str] tags: List of tags
        :param list[object] datasets: List of input/output datasets

        :returns: Job handle
        :rtype: str
        """

        if not build_only:
            build_only = None

        job = Job(
            machine_type=machine_type,
            container=container,
            project_id=project_id,
            name=name,
            command=command,
            ports=ports,
            is_public=is_public,
            working_directory=working_directory,
            experiment_id=experiment_id,
            env_vars=env_vars,
            use_dockerfile=use_dockerfile,
            is_preemptible=is_preemptible,
            project=project,
            started_by_user_id=started_by_user_id,
            rel_dockerfile_path=rel_dockerfile_path,
            registry_username=registry_username,
            registry_password=registry_password,
            cluster=cluster,
            cluster_id=cluster_id,
            target_node_attrs=node_attrs,
            workspace_file_name=workspace_file_name,
            registry_target=registry_target,
            registry_target_username=registry_target_username,
            registry_target_password=registry_target_password,
            build_only=build_only,
            datasets=datasets,
        )
        repository = self.build_repository(CreateJob)
        handle = repository.create(job, data=data)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)

        return handle

    def delete(self, job_id):
        """
        Method to remove job.

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            job_client.delete(
                job_id='Your_job_id_here'
            )

        :param str job_id: id of job that you want to remove
        :raises: exceptions.GradientSdkError
        """
        repository = self.build_repository(DeleteJob)
        repository.delete(job_id)

    def stop(self, job_id):
        """
        Method to stop working job

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            job_client.stop(
                job_id='Your_job_id_here'
            )

        :param job_id: id of job that we want to stop
        :raises: exceptions.GradientSdkError
        """
        repository = self.build_repository(StopJob)
        repository.stop(job_id)

    def list(self, project_id=None, project=None, experiment_id=None, tags=None):
        """
        Method to list jobs.

        To retrieve all user jobs:

        .. code-block:: python
            :linenos:

            jobs = job_client.list()

        To list jobs from project:

        .. code-block:: python
            :linenos:

            job = job_client.list(
                project_id="Your_project_id_here",
            )

        :param str project_id: id of project that you want to list jobs
        :param str project: name of project that you want to list jobs
        :param str experiment_id: id of experiment that you want to list jobs
        :param list[str]|tuple[str] tags: tags to filter jobs with OR

        :returns: list of job models
        :rtype: list[Job]
        """
        repository = self.build_repository(ListJobs)
        jobs = repository.list(
            project_id=project_id,
            project=project,
            experiment_id=experiment_id,
            tags=tags,
        )
        return jobs

    def logs(self, job_id, line=1, limit=10000):
        """
        Method to retrieve job logs.

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            job_logs = job_client.logs(
                job_id='Your_job_id_here',
                line=100,
                limit=100
            )

        :param str job_id: id of job that we want to retrieve logs
        :param int line: from what line you want to retrieve logs. Default 0
        :param int limit: how much lines you want to retrieve logs. Default 10000

        :returns: list of formatted logs lines
        :rtype: list
        """
        repository = self.build_repository(ListJobLogs)
        logs = repository.list(id=job_id, line=line, limit=limit)
        return logs

    def yield_logs(self, job_id, line=1, limit=10000):
        """Get log generator. Polls the API for new logs

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            job_logs_generator = job_client.yield_logs(
                job_id='Your_job_id_here',
                line=100,
                limit=100
            )

        :param str job_id:
        :param int line: line number at which logs starts to display on screen
        :param int limit: maximum lines displayed on screen, default set to 10 000

        :returns: generator yielding LogRow instances
        :rtype: Iterator[models.LogRow]
        """

        repository = self.build_repository(ListJobLogs)
        logs = repository.yield_logs(id=job_id, line=line, limit=limit)
        return logs

    def artifacts_delete(self, job_id, files=None):
        """
        Method to delete job artifact.

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            job_client.artifacts_delete(
                job_id='Your_job_id_here',
                files=files,
            )

        :param str job_id: Id of job which artifact you want to delete
        :param str files: if you wish to remove only few files from artifact pass string with names of this files
            separated by comma ``,``

        :raises: exceptions.GradientSdkError
        """
        repository = self.build_repository(DeleteJobArtifacts)
        repository.delete(id_=job_id, files=files)

    def artifacts_get(self, job_id):
        """
        Method to retrieve federated access information for job artifacts.

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            artifacts = job_client.artifacts_get(
                job_id='your_job_id_here',
            )

        :param job_id: Id of job from which you want to retrieve artifacts information about location

        :returns: Information about artifact place
        :rtype: dict
        """
        repository = self.build_repository(GetJobArtifacts)
        data = repository.get(jobId=job_id)
        return data

    def artifacts_list(self, job_id, files=None, size=False, links=True, start_after=None):
        """
        Method to retrieve all artifacts files.

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            artifacts = job_client.artifacts_list(
                job_id='your_job_id_here',
                files='your_files,here',
                size=False,
                links=True,
                start_after='key',
            )

        :param str job_id: to limit artifact from this job.
        :param str files: to limit result only to file names provided. You can use wildcard option ``*``.
        :param bool size: flag to show file size. Default value is set to False.
        :param bool links: flag to show file url. Default value is set to True.
        :params str start_after: key to list after

        :returns: list of files with description if specified from job artifacts.
        :rtype: Pagination
        """
        repository = self.build_repository(ListJobArtifacts)
        return repository.list(jobId=job_id, files=files, links=links, size=size, start_after=start_after)

    def get_metrics(self, job_id, start=None, end=None, interval="30s", built_in_metrics=None):
        """Get job metrics

        :param str job_id: ID of a job
        :param datetime.datetime|str start:
        :param datetime.datetime|str end:
        :param str interval:
        :param list[str] built_in_metrics: List of metrics to get if different than default
                    Available builtin metrics: cpuPercentage, memoryUsage, gpuMemoryFree, gpuMemoryUsed, gpuPowerDraw,
                                            gpuTemp, gpuUtilization, gpuMemoryUtilization

        :returns: Metrics of a job
        :rtype: dict[str,dict[str,list[dict]]]
        """

        repository = self.build_repository(GetJobMetrics)
        metrics = repository.get(
            id=job_id,
            start=start,
            end=end,
            interval=interval,
            built_in_metrics=built_in_metrics,
        )
        return metrics

    def list_metrics(self, job_id, start=None, end=None, interval="30s"):
        """List job metrics

        :param str job_id: ID of a job
        :param datetime.datetime|str start:
        :param datetime.datetime|str end:
        :param str interval:
        :returns: Metrics of a job
        :rtype: dict[str,dict[str,list[dict]]]
        """

        repository = self.build_repository(ListJobMetrics)
        metrics = repository.get(
            id=job_id,
            start=start,
            end=end,
            interval=interval,
        )
        return metrics

    def stream_metrics(self, job_id, interval="30s", built_in_metrics=None):
        """Stream live job metrics

        :param str job_id: ID of a job
        :param str interval:
        :param list[str] built_in_metrics: List of metrics to get if different than default
                    Available builtin metrics: cpuPercentage, memoryUsage, gpuMemoryFree, gpuMemoryUsed, gpuPowerDraw,
                                            gpuTemp, gpuUtilization, gpuMemoryUtilization

        :returns: Generator object yielding live job metrics
        :rtype: Iterable[dict]
        """

        repository = self.build_repository(StreamJobMetrics)
        metrics = repository.stream(
            id=job_id,
            interval=interval,
            built_in_metrics=built_in_metrics,
        )
        return metrics



from .common import AlterResource, CreateResource, DeleteResource, GetResource, ListResources
from .. import config, serializers


class DatasetMixin(object):
    SERIALIZER_CLS = serializers.DatasetSchema

    @staticmethod
    def _get_api_url(**kwargs):
        return config.config.CONFIG_HOST

    @staticmethod
    def get_request_url(**kwargs):
        url = "/datasets"
        if kwargs.get("id"):
            url += "/{}".format(kwargs["id"])
        return url


class ListDatasets(DatasetMixin, ListResources):
    def _get_request_params(self, kwargs):
        limit = kwargs.get("limit") or 20
        offset = kwargs.get("offset") or 0
        return {
            "filter[limit]": limit,
            "filter[skip]": offset,
            "filter[order][]": "name ASC",
        }


class CreateDataset(DatasetMixin, CreateResource):
    HANDLE_FIELD = "id"


class GetDataset(DatasetMixin, GetResource):
    pass


class UpdateDataset(DatasetMixin, AlterResource):
    def _get_request_json(self, kwargs):
        data = dict(kwargs)
        if "id" in data:
            del data["id"]
        return data


class DeleteDataset(DatasetMixin, DeleteResource):
    pass


class GetDatasetRef(DatasetMixin, GetResource):
    SERIALIZER_CLS = serializers.DatasetRefSchema

    @staticmethod
    def get_request_url(**kwargs):
        return "/datasets/ref/{}".format(kwargs["id"])



import attr


@attr.s
class StorageProvider(object):
    """
    Storage provider class
    """
    id = attr.ib(type=str, default=None)
    type = attr.ib(type=str, default=None)
    name = attr.ib(type=str, default=None)
    config = attr.ib(type=dict, factory=dict)



from .base_client import BaseClient
from .. import models, repositories


class StorageProvidersClient(BaseClient):

    def list(self, limit=20, offset=0):
        """Get list of your storage providers

        :param int limit: limit of the number results
        :param int offset: skip number of results

        :returns: list of storage providers
        :rtype: list[models.StorageProvider]
        """

        repository = self.build_repository(repositories.ListStorageProviders)
        return repository.list(limit=limit, offset=offset)

    def delete(self, storage_provider_id):
        """Delete a storage provider

        :param int storage_provider_id: ID [required]

        :returns:
        :rtype: None
        """

        repository = self.build_repository(repositories.DeleteStorageProvider)
        repository.delete(storage_provider_id)

    def get(self, storage_provider_id):
        """Delete a storage provider

        :param int storage_provider_id: ID [required]

        :returns: storage provider
        :rtype: models.StorageProvider
        """
        repository = self.build_repository(repositories.GetStorageProvider)
        return repository.get(id=storage_provider_id)

    def create_s3(self, name, bucket, access_key, secret_access_key, endpoint=None, region=None,
                  signature_version=None):
        """Create a new S3 storage provider

        :param str name: Name of new AWS storage provider [required]
        :param str bucket: S3 bucket [required]
        :param str access_key: S3 access key ID [required]
        :param str secret_access_key: S3 access key ID [required]
        :param str endpoint: S3 endpoint URL
        :param str region: S3 region
        :param str signature_version: S3 signature version (ex: v4)

        :returns: storage provider ID
        :rtype: str
        """

        config = {
            'bucket': bucket,
            'accessKey': access_key,
            'secretAccessKey': secret_access_key,
        }
        if endpoint:
            config['endpoint'] = endpoint
        if region:
            config['region'] = region
        if signature_version:
            config['signatureVersion'] = signature_version

        storage_provider = models.StorageProvider(
            type='s3',
            name=name,
            config=config,
        )

        repository = self.build_repository(repositories.CreateStorageProvider)
        return repository.create(storage_provider)

    def update_s3(self, storage_provider_id, name=None, bucket=None, access_key=None, secret_access_key=None,
                  endpoint=None, region=None, signature_version=None):
        """Update an existing S3 storage provider

        :param str storage_provider_id: Storage provider ID
        :param str name: Storage provider name
        :param str bucket: S3 bucket
        :param str access_key: S3 access key ID
        :param str secret_access_key: S3 access key ID
        :param str endpoint: S3 endpoint URL
        :param str region: S3 region
        :param str signature_version: S3 signature version (ex: v4)

        :returns:
        :rtype: None
        """

        config = {}
        if bucket:
            config['bucket'] = bucket
        if access_key:
            config['accessKey'] = access_key
        if secret_access_key:
            config['secretAccessKey'] = secret_access_key
        if endpoint is not None:
            config['endpoint'] = endpoint
        if region is not None:
            config['region'] = region
        if signature_version is not None:
            config['signatureVersion'] = signature_version

        storage_provider = models.StorageProvider(
            name=name,
            config=config,
        )

        repository = self.build_repository(repositories.UpdateStorageProvider)
        return repository.update(storage_provider_id, storage_provider)



import abc
import collections
import datetime
import json

import dateutil
import six
import websocket

from .. import serializers, sdk_exceptions
from ..clients import http_client
from ..config import config
from ..sdk_exceptions import ResourceFetchingError, ResourceCreatingDataError, ResourceCreatingError, GradientSdkError
from ..utils import MessageExtractor, concatenate_urls


@six.add_metaclass(abc.ABCMeta)
class BaseRepository(object):
    VALIDATION_ERROR_MESSAGE = "Failed to fetch data"

    def __init__(self, api_key, logger, ps_client_name=None):
        self.api_key = api_key
        self.logger = logger
        self.ps_client_name = ps_client_name

    @abc.abstractmethod
    def get_request_url(self, **kwargs):
        """Get url to the endpoint (without base api url)

        :rtype: str
        """
        pass

    @abc.abstractmethod
    def _get_api_url(self, **kwargs):
        """Get base url to the api

        :rtype: str
        """
        pass

    def _get_client(self, **kwargs):
        """
        :rtype: http_client.API
        """
        api_url = self._get_api_url(**kwargs)
        client = http_client.API(
            api_url=api_url,
            api_key=self.api_key,
            logger=self.logger,
            ps_client_name=self.ps_client_name,
        )
        return client

    def _get(self, **kwargs):
        json_ = self._get_request_json(kwargs)
        params = self._get_request_params(kwargs)
        url = self.get_request_url(**kwargs)
        client = self._get_client(**kwargs)
        response = self._send_request(client, url, json=json_, params=params)
        gradient_response = http_client.GradientResponse.interpret_response(response)

        return gradient_response

    def _send_request(self, client, url, json=None, params=None):
        response = client.get(url, json=json, params=params)
        return response

    def _validate_response(self, response):
        if not response.ok:
            msg = self.VALIDATION_ERROR_MESSAGE
            errors = MessageExtractor().get_message_from_response_data(response.data)
            if errors:
                msg += ": " + errors
            raise ResourceFetchingError(msg)

    def _get_request_json(self, kwargs):
        return None

    def _get_request_params(self, kwargs):
        return None


@six.add_metaclass(abc.ABCMeta)
class ListResources(BaseRepository):
    SERIALIZER_CLS = None

    def _parse_objects(self, data, **kwargs):
        instances = []
        instance_dicts = self._get_instance_dicts(data, **kwargs)
        for instance_dict in instance_dicts:
            instance = self._parse_object(instance_dict)
            instances.append(instance)

        return instances

    def _get_instance_dicts(self, data, **kwargs):
        return data

    def _get_meta_data(self, resp):
        pass

    def _parse_object(self, instance_dict):
        """
        :param dict instance_dict:
        :return: model instance
        """
        instance = self.SERIALIZER_CLS().get_instance(instance_dict)
        return instance

    def list(self, **kwargs):
        response = self._get(**kwargs)
        self._validate_response(response)
        instances = self._get_instances(response, **kwargs)
        if kwargs.get("get_meta"):
            # Added for now to prevent unnecessary action to update all list commands
            meta_data = self._get_meta_data(response)
            return instances, meta_data
        return instances

    def _get_instances(self, response, **kwargs):
        if not response.data:
            return []

        objects = self._parse_objects(response.data, **kwargs)
        return objects


@six.add_metaclass(abc.ABCMeta)
class GetResource(BaseRepository):
    SERIALIZER_CLS = None

    def _parse_object(self, instance_dict, **kwargs):
        """
        :param dict instance_dict:
        :return: model instance
        """
        instance = self.SERIALIZER_CLS().get_instance(instance_dict)
        return instance

    def get(self, **kwargs):
        response = self._get(**kwargs)
        self._validate_response(response)
        instance = self._get_instance(response, **kwargs)
        return instance

    def _get_instance(self, response, **kwargs):
        try:
            objects = self._parse_object(response.data, **kwargs)
        except GradientSdkError:
            raise
        except Exception:
            msg = "Error parsing response data: {}".format(str(response.body))
            raise ResourceFetchingError(msg)

        return objects


@six.add_metaclass(abc.ABCMeta)
class CreateResource(BaseRepository):
    SERIALIZER_CLS = None
    VALIDATION_ERROR_MESSAGE = "Failed to create resource"
    HANDLE_FIELD = "handle"

    def create(self, instance, data=None, path=None):
        instance_dict = self._get_instance_dict(instance)
        response = self._send_create_request(instance_dict, data=data, path=path)
        self._validate_response(response)
        handle = self._process_response(response)
        return handle

    def _get_instance_dict(self, instance):
        serializer = self._get_serializer()
        serialization_result = serializer.dump(instance)
        instance_dict = serialization_result.data
        if serialization_result.errors:
            raise ResourceCreatingDataError(str(serialization_result.errors))

        instance_dict = self._process_instance_dict(instance_dict)
        return instance_dict

    def _get_serializer(self):
        serializer = self.SERIALIZER_CLS()
        return serializer

    def _send_create_request(self, instance_dict, data=None, path=None):
        url = self.get_request_url(**instance_dict)
        client = self._get_client(**instance_dict)
        json_ = self._get_request_json(instance_dict)
        params = self._get_request_params(instance_dict)
        files = self._get_request_files(path)
        response = client.post(url, params=params, json=json_, data=data, files=files)
        gradient_response = http_client.GradientResponse.interpret_response(response)
        return gradient_response

    def _process_response(self, response):
        try:
            return self._get_id_from_response(response)
        except Exception as e:
            raise ResourceCreatingError(e)

    def _get_id_from_response(self, response):
        handle = response.data[self.HANDLE_FIELD]
        return handle

    def _process_instance_dict(self, instance_dict):
        return instance_dict

    def _get_request_json(self, instance_dict):
        return instance_dict

    def _get_request_files(self, path):
        return None


@six.add_metaclass(abc.ABCMeta)
class AlterResource(BaseRepository):
    SERIALIZER_CLS = None

    def update(self, id, instance):
        instance_dict = self._get_instance_dict(instance)
        self._run(id=id, **instance_dict)

    def _get_instance_dict(self, instance):
        serializer = self._get_serializer()
        serialization_result = serializer.dump(instance)
        instance_dict = serialization_result.data
        return instance_dict

    def _get_serializer(self):
        serializer = self.SERIALIZER_CLS()
        return serializer

    def _run(self, **kwargs):
        url = self.get_request_url(**kwargs)
        response = self._send(url, **kwargs)
        self._validate_response(response)
        return response

    def _send(self, url, **kwargs):
        client = self._get_client(**kwargs)
        json_data = self._get_request_json(kwargs)
        response = self._send_request(client, url, json_data=json_data)
        gradient_response = http_client.GradientResponse.interpret_response(response)
        return gradient_response

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response


@six.add_metaclass(abc.ABCMeta)
class DeleteResource(AlterResource):
    VALIDATION_ERROR_MESSAGE = "Failed to delete resource"

    def delete(self, id_, **kwargs):
        self._run(id=id_, **kwargs)

    def _send_request(self, client, url, json_data=None):
        response = client.delete(url, json=json_data)
        return response


@six.add_metaclass(abc.ABCMeta)
class StartResource(AlterResource):
    VALIDATION_ERROR_MESSAGE = "Unable to start instance"

    def start(self, id_):
        self._run(id=id_)

    def _send_request(self, client, url, json_data=None):
        response = client.put(url, json=json_data)
        return response


@six.add_metaclass(abc.ABCMeta)
class StopResource(AlterResource):
    VALIDATION_ERROR_MESSAGE = "Unable to stop instance"

    def stop(self, id_):
        self._run(id=id_)

    def _send_request(self, client, url, json_data=None):
        response = client.put(url, json=json_data)
        return response


@six.add_metaclass(abc.ABCMeta)
class GetMetrics(GetResource):
    OBJECT_TYPE = None

    DEFAULT_INTERVAL = "30s"
    DEFAULT_METRICS = ["cpuPercentage", "memoryUsage"]

    @abc.abstractmethod
    def _get_instance_by_id(self, instance_id, **kwargs):
        pass

    def _get_metrics_api_url(self, instance, protocol="https"):
        if not instance.metrics_url:
            raise GradientSdkError("Metrics API url not found")

        metrics_api_url = concatenate_urls(protocol + "://", instance.metrics_url)
        return metrics_api_url

    def _get(self, **kwargs):
        new_kwargs = self._get_kwargs(kwargs)
        rv = super(GetMetrics, self)._get(**new_kwargs)
        return rv

    def _get_kwargs(self, kwargs):
        instance_id = kwargs["id"]
        built_in_metrics = self._get_built_in_metrics_comma_separated(kwargs)
        instance = self._get_instance_by_id(instance_id)
        started_date = self._get_start_date(instance, kwargs)
        end = self._get_end_date(instance, kwargs)
        interval = kwargs.get("interval") or self.DEFAULT_INTERVAL
        metrics_api_url = self._get_metrics_api_url(instance)
        new_kwargs = {
            "charts": built_in_metrics,
            "start": started_date,
            "interval": interval,
            "objecttype": self.OBJECT_TYPE,
            "handle": instance_id,
            "metrics_api_url": metrics_api_url,
        }
        if end:
            new_kwargs["end"] = end

        return new_kwargs

    def get_request_url(self, **kwargs):
        return "metrics/api/v1/range"

    def _get_api_url(self, **kwargs):
        api_url = kwargs["metrics_api_url"]
        return api_url

    def _get_built_in_metrics_comma_separated(self, kwargs):
        metrics_list = self._get_built_in_metrics_list(kwargs)
        metrics_list = ",".join(metrics_list)
        return metrics_list

    def _get_built_in_metrics_list(self, kwargs):
        metrics = kwargs.get("built_in_metrics") or self.DEFAULT_METRICS
        return metrics

    def _get_start_date(self, instance, kwargs):
        datetime_string = kwargs.get("start") or instance.dt_started or instance.dt_created
        if not datetime_string:
            return None

        datetime_string = self._format_datetime(datetime_string)
        return datetime_string

    def _get_end_date(self, instance, kwargs):
        datetime_string = kwargs.get("end")
        if not datetime_string:
            return None

        datetime_string = self._format_datetime(datetime_string)
        return datetime_string

    def _get_request_params(self, kwargs):
        params = kwargs.copy()
        params.pop("metrics_api_url", None)
        return params

    def _parse_object(self, instance_dict, **kwargs):
        charts = instance_dict["charts"]
        return charts

    def _format_datetime(self, some_datetime):
        if not isinstance(some_datetime, datetime.datetime):
            some_datetime = dateutil.parser.parse(some_datetime)

        datetime_str = some_datetime.strftime("%Y-%m-%dT%H:%M:%SZ")
        return datetime_str

class ListMetrics(GetResource):
    OBJECT_TYPE = None

    DEFAULT_INTERVAL = "30s"

    @abc.abstractmethod
    def _get_instance_by_id(self, instance_id, **kwargs):
        pass

    def _get_metrics_api_url(self, instance, protocol="https"):
        if not instance.metrics_url:
            raise GradientSdkError("Metrics API url not found")

        metrics_api_url = concatenate_urls(protocol + "://", instance.metrics_url)
        return metrics_api_url

    def _get(self, **kwargs):
        new_kwargs = self._get_kwargs(kwargs)
        rv = super(ListMetrics, self)._get(**new_kwargs)
        return rv

    def _get_kwargs(self, kwargs):
        instance_id = kwargs["id"]
        instance = self._get_instance_by_id(instance_id)
        started_date = self._get_start_date(instance, kwargs)
        end = self._get_end_date(instance, kwargs)
        interval = kwargs.get("interval") or self.DEFAULT_INTERVAL
        metrics_api_url = self._get_metrics_api_url(instance)
        new_kwargs = {
            "start": started_date,
            "interval": interval,
            "objecttype": self.OBJECT_TYPE,
            "handle": instance_id,
            "metrics_api_url": metrics_api_url,
        }
        if end:
            new_kwargs["end"] = end

        return new_kwargs

    def get_request_url(self, **kwargs):
        return "metrics/api/v1/list"

    def _get_api_url(self, **kwargs):
        api_url = kwargs["metrics_api_url"]
        return api_url

    def _get_start_date(self, instance, kwargs):
        datetime_string = kwargs.get("start") or instance.dt_started or instance.dt_created
        if not datetime_string:
            return None

        datetime_string = self._format_datetime(datetime_string)
        return datetime_string

    def _get_end_date(self, instance, kwargs):
        datetime_string = kwargs.get("end")
        if not datetime_string:
            return None

        datetime_string = self._format_datetime(datetime_string)
        return datetime_string

    def _get_request_params(self, kwargs):
        params = kwargs.copy()
        params.pop("metrics_api_url", None)
        return params

    def _parse_object(self, instance_dict, **kwargs):
        chart_names = instance_dict["chart_names"]
        return chart_names

    def _format_datetime(self, some_datetime):
        if not isinstance(some_datetime, datetime.datetime):
            some_datetime = dateutil.parser.parse(some_datetime)

        datetime_str = some_datetime.strftime("%Y-%m-%dT%H:%M:%SZ")
        return datetime_str

@six.add_metaclass(abc.ABCMeta)
class StreamMetrics(BaseRepository):
    OBJECT_TYPE = None

    DEFAULT_INTERVAL = "30s"
    DEFAULT_METRICS = ["cpuPercentage", "memoryUsage"]

    def _get_metrics_api_url(self, instance, protocol="https"):
        if not instance.metrics_url:
            raise GradientSdkError("Metrics API url not found")

        metrics_api_url = concatenate_urls(protocol + "://", instance.metrics_url)
        return metrics_api_url

    def _get_api_url(self, **kwargs):
        api_url = kwargs["metrics_api_url"]
        return api_url

    def stream(self, **kwargs):
        while True:
            try:
                connection = self._get_connection(kwargs)
                self._send_chart_descriptor(connection, kwargs)
                stream_generator = self._get_stream_generator(connection)
                for data in stream_generator:
                    self.logger.debug("Metrics data: {}".format(data))
                    yield data
            except websocket.WebSocketConnectionClosedException as e:
                self.logger.debug("WebSocketConnectionClosedException: {}".format(e))
            except sdk_exceptions.EndWebsocketStream:
                return

    def _get_connection(self, kwargs):
        url = self._get_full_url(kwargs)
        self.logger.debug("(Re)opening websocket connection to: {}".format(url))
        ws = websocket.create_connection(url)
        self.logger.debug("Connected")
        return ws

    def _get_full_url(self, kwargs):
        instance_id = kwargs["id"]
        metrics_api_url = self._get_metrics_api_url(instance_id, protocol="wss")
        url = concatenate_urls(metrics_api_url, self.get_request_url())
        return url

    def get_request_url(self, **kwargs):
        return "metrics/api/v1/stream"

    def _get_chart_descriptor(self, kwargs):
        instance_id = kwargs["id"]
        built_in_metrics = self._get_built_in_metrics_list(kwargs)
        interval = kwargs.get("interval") or self.DEFAULT_INTERVAL
        descriptor_json = collections.OrderedDict(
            (
                ("chart_names", built_in_metrics),
                ("handles", [instance_id]),
                ("object_type", self.OBJECT_TYPE),
                ("poll_interval", interval),
            )
        )

        descriptor = json.dumps(descriptor_json)

        return descriptor

    def _get_built_in_metrics_list(self, kwargs):
        metrics = kwargs.get("built_in_metrics") or self.DEFAULT_METRICS
        return metrics

    def _send_chart_descriptor(self, connection, kwargs):
        descriptor = self._get_chart_descriptor(kwargs)
        self.logger.debug("Sending chart descriptor: {}".format(descriptor))
        response = connection.send(descriptor)
        self.logger.debug("Chart descriptor sent. Response: {}".format(response))

    def _get_stream_generator(self, connection):
        return connection

class ListLogs(ListResources):
    @abc.abstractmethod
    def _get_request_params(self, kwargs):
        pass

    def _get_api_url(self, **kwargs):
        return config.CONFIG_LOG_HOST

    def get_request_url(self, **kwargs):
        return "/jobs/logs"

    def yield_logs(self, id, line=1, limit=10000):

        gen = self._get_logs_generator(id, line, limit)
        return gen

    def _get_logs_generator(self, id, line, limit):
        while True:
            logs = self.list(id=id, line=line, limit=limit)

            for log in logs:
                # stop generator - "PSEOF" indicates there are no more logs
                if log.message == "PSEOF":
                    return

                yield log
                line += 1

    def _parse_objects(self, log_rows, **kwargs):
        serializer = serializers.LogRowSchema()
        log_rows = [serializer.get_instance(row) for row in log_rows]
        return log_rows



from .base_client import BaseClient
from .. import models, repositories


class DatasetVersionsClient(BaseClient):

    def list(self, dataset_id, is_committed=True, limit=20, offset=0):
        """Get list of your dataset versions

        :param str dataset_id: Dataset ID [required]
        :param bool is_committed: Filter versions by commit status
        :param int limit: Limit results
        :param int offset: Skip results

        :returns: List of dataset versions
        :rtype: list[models.DatasetVersion]
        """

        repository = self.build_repository(repositories.ListDatasetVersions)
        return repository.list(id=dataset_id, is_committed=is_committed, limit=limit, offset=offset)

    def delete(self, dataset_version_id):
        """Delete a dataset version

        :param str dataset_version_id: Dataset version ID (ex: dataset_id:version) [required]

        :returns:
        :rtype: None
        """

        repository = self.build_repository(repositories.DeleteDatasetVersion)
        repository.delete(dataset_version_id)

    def get(self, dataset_version_id):
        """Delete a dataset version

        :param str dataset_version_id: Dataset version ID (ex: dataset_id:version) [required]

        :returns: dataset
        :rtype: models.DatasetVersion
        """
        repository = self.build_repository(repositories.GetDatasetVersion)
        return repository.get(id=dataset_version_id)

    def create(self, dataset_id, message=None):
        """Create a new dataset version

        :param str id: Dataset ID [required]
        :param str message: Dataset version message

        :returns: dataset ID
        :rtype: str
        """

        dataset_version = models.DatasetVersion(
            dataset_id=dataset_id,
            message=message,
        )

        repository = self.build_repository(repositories.CreateDatasetVersion)
        return repository.create(dataset_version)

    def update(self, dataset_version_id, message=None, is_committed=None):
        """Update an existing S3 dataset

        :param str dataset_version_id: Dataset version ID (ex: dataset_id:version)
        :param str message: Dataset version message
        :param bool is_committed: Mark dataset version as committed

        :returns:
        :rtype: None
        """

        dataset = models.DatasetVersion(
            message=message,
            is_committed=is_committed,
        )

        repository = self.build_repository(repositories.UpdateDatasetVersion)
        return repository.update(dataset_version_id, dataset)

    def generate_pre_signed_s3_url(self, dataset_version_id, method, params=None):
        """Generate pre-signed URL for S3 storage providers

        :param str dataset_version_id: Dataset version ID (ex: dataset_id:version)
        :param str method: S3 method
        :param dict params: S3 params

        :returns:
        :rtype: DatasetVersionPreSignedURL
        """

        call = {'method': method}
        if params:
            call['params'] = params

        repository = self.build_repository(repositories.GenerateDatasetVersionPreSignedS3Urls)
        results = repository.generate(dataset_version_id, [call])
        return results[0]

    def generate_pre_signed_s3_urls(self, dataset_version_id, calls):
        """Generate pre-signed URLs for S3 storage providers

        :param str dataset_version_id: Dataset version ID (ex: dataset_id:version)
        :param list[dict] calls: List of S3 calls

        :returns:
        :rtype: list[models.DatasetVersionPreSignedURL]
        """

        repository = self.build_repository(repositories.GenerateDatasetVersionPreSignedS3Urls)
        return repository.generate(dataset_version_id, calls)



import mock
from click.testing import CliRunner

from gradient.api_sdk.clients.http_client import default_headers
from gradient.cli import cli
from tests import MockResponse, example_responses

EXPECTED_HEADERS = default_headers.copy()
EXPECTED_HEADERS["ps_client_name"] = "gradient-cli"

EXPECTED_HEADERS_WITH_CHANGED_API_KEY = EXPECTED_HEADERS.copy()
EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"


class TestTensorboardsCreate(object):
    URL = "https://services.paperspace.io/tensorboards/v1/"
    COMMAND = [
        "tensorboards",
        "create",
        "--experiment", "some_experiment_id",
        "--experiment", "some_other_experiment_id",
    ]
    EXPECTED_REQUEST_JSON = {"experiments": ["some_experiment_id", "some_other_experiment_id"]}
    EXPECTED_RESPONSE_JSON = example_responses.TENSORBOARD_CREATE_RESPONSE_JSON
    EXPECTED_STDOUT = """Created new tensorboard with id: some_id\n"""

    COMMAND_WITH_API_KEY_CHANGED = [
        "tensorboards",
        "create",
        "--experiment", "some_experiment_id",
        "--experiment", "some_other_experiment_id",
        "--apiKey", "some_key",
    ]
    COMMAND_WITH_ALL_OPTIONS = [
        "tensorboards",
        "create",
        "--experiment", "some_experiment_id",
        "--experiment", "some_other_experiment_id",
        "--image", "some_image",
        "--username", "some_username",
        "--password", "some_password",
        "--apiKey", "some_key",
    ]

    EXPECTED_REQUEST_JSON_WITH_ALL_OPTIONS = {
        "experiments": [
            "some_experiment_id",
            "some_other_experiment_id",
        ],
        "image": "some_image",
        "username": "some_username",
        "password": "some_password"
    }
    COMMAND_WITH_OPTIONS_FILE_USED = ["tensorboards", "create", "--optionsFile", ]  # path added in test

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"title": "Invalid credentials provided"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to create resource: Invalid credentials provided\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_request_when_command_was_executed_with_required_options(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             data=None,
                                             files=None,
                                             params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_request_with_changed_api_key_when_api_key_option_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_CHANGED)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             data=None,
                                             files=None,
                                             params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_request_when_command_was_executed_with_all_options(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_ALL_OPTIONS)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON_WITH_ALL_OPTIONS,
                                             data=None,
                                             files=None,
                                             params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, tensorboards_create_config_path):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE_USED[:] + [tensorboards_create_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON_WITH_ALL_OPTIONS,
                                             data=None,
                                             files=None,
                                             params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"


class TestTensorboardsDetail(object):
    URL = "https://services.paperspace.io/tensorboards/v1/some_id/"
    COMMAND = [
        "tensorboards",
        "details",
        "--id", "some_id",
    ]
    EXPECTED_RESPONSE_JSON = example_responses.TENSORBOARD_DETAIL_RESPONSE_JSON
    EXPECTED_STDOUT = """+-------+--------------------------------------------------------+
| ID    | some_id                                                |
+-------+--------------------------------------------------------+
| Image | tensorflow/tensorflow:latest-py3                       |
| URL   | https://aws-testing.paperspace.io/tensorboard/some_id/ |
| State | Some State                                             |
+-------+--------------------------------------------------------+
+--------------------------+------------+
| Experiments ID           | State      |
+--------------------------+------------+
| some_experiment_id       | Some State |
| some_other_experiment_id | Some State |
+--------------------------+------------+
"""
    # TODO later change response to contain information about state: | State          | 1                             |

    COMMAND_WITH_API_KEY_CHANGED = [
        "tensorboards",
        "details",
        "--id", "some_id",
        "--apiKey", "some_key",
    ]
    COMMAND_WITH_OPTIONS_FILE_USED = ["tensorboards", "details", "--optionsFile", ]  # path added in test

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"title": "Invalid credentials provided"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to create resource: Invalid credentials provided\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_request_when_command_was_executed_with_required_options(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=None,
                                             params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_request_with_changed_api_key_when_api_key_option_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_CHANGED)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=None,
                                             params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, post_patched, tensorboards_details_config_path):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE_USED[:] + [tensorboards_details_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=None,
                                             params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"


class TestTensorboardsList(object):
    URL = "https://services.paperspace.io/tensorboards/v1/"
    COMMAND = ["tensorboards", "list"]
    EXPECTED_RESPONSE_JSON = example_responses.TENSORBOARD_LIST_RESPONSE_JSON
    EXPECTED_STDOUT = """+-----------------+----------------------------------------------------------------+------------+
| ID              | URL                                                            | STATE      |
+-----------------+----------------------------------------------------------------+------------+
| tbrs2kcjman4ly  | https://aws-testing.paperspace.io/tensorboard/tbrs2kcjman4ly/  | Some State |
| tbskzep6d9po04d | https://aws-testing.paperspace.io/tensorboard/tbskzep6d9po04d/ | Some State |
| tbsaq6hggzxcnet | https://aws-testing.paperspace.io/tensorboard/tbsaq6hggzxcnet/ | Some State |
| tbwuzalec7ik58  | https://aws-testing.paperspace.io/tensorboard/tbwuzalec7ik58/  | Some State |
+-----------------+----------------------------------------------------------------+------------+
"""

    COMMAND_WITH_API_KEY_CHANGED = ["tensorboards", "list", "--apiKey", "some_key"]
    COMMAND_WITH_OPTIONS_FILE_USED = ["tensorboards", "list", "--optionsFile", ]  # path added in test

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"title": "Invalid credentials provided"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to create resource: Invalid credentials provided\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_request_when_command_was_executed_with_required_options(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=None,
                                             params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_request_with_changed_api_key_when_api_key_option_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_CHANGED)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=None,
                                             params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, post_patched, tensorboards_details_config_path):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE_USED[:] + [tensorboards_details_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=None,
                                             params=None)
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"


class TestTensorboardsAddExperiment(object):
    URL = "https://services.paperspace.io/tensorboards/v1/some_id"
    COMMAND = ["tensorboards", "add-experiments", "--id", "some_id", "--experiment", "some_third_experiment_id"]

    EXPECTED_RESPONSE_JSON = example_responses.TENSORBOARD_UPDATE_RESPONSE_JSON
    EXPECTED_REQUEST_JSON = {"added_experiments": ["some_third_experiment_id"], "removed_experiments": []}
    EXPECTED_STDOUT = """+-------+--------------------------------------------------------+
| ID    | some_id                                                |
+-------+--------------------------------------------------------+
| Image | tensorflow/tensorflow:latest-py3                       |
| URL   | https://aws-testing.paperspace.io/tensorboard/some_id/ |
| State | Some State                                             |
+-------+--------------------------------------------------------+
+--------------------------+------------+
| Experiments ID           | State      |
+--------------------------+------------+
| some_experiment_id       | Some State |
| some_other_experiment_id | Some State |
| some_third_experiment_id | Some State |
+--------------------------+------------+
"""

    COMMAND_WITH_API_KEY_CHANGED = [
        "tensorboards", "add-experiments", "--id", "some_id",
        "--experiment", "some_third_experiment_id", "--apiKey", "some_key"
    ]

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"title": "Invalid credentials provided"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to update resource: Invalid credentials provided\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_send_valid_request_when_command_was_executed_with_required_options(self, put_patched):
        put_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        put_patched.assert_called_once_with(
            self.URL,
            headers=EXPECTED_HEADERS,
            json=self.EXPECTED_REQUEST_JSON,
            params=None,
            data=None,
        )
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_send_request_with_changed_api_key_when_api_key_option_was_used(self, put_patched):
        put_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=401)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_CHANGED)

        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN, result.exc_info
        put_patched.assert_called_once_with(
            self.URL,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
            json=self.EXPECTED_REQUEST_JSON,
            params=None,
            data=None,
        )


class TestTensorboardsRemoveExperiment(object):
    URL = "https://services.paperspace.io/tensorboards/v1/some_id"
    COMMAND = ["tensorboards", "remove-experiments", "--id", "some_id", "--experiment", "some_other_experiment_id"]

    EXPECTED_RESPONSE_JSON = example_responses.TENSORBOARD_UPDATE_REMOVE_RESPONSE_JSON
    EXPECTED_REQUEST_JSON = {"added_experiments": [], "removed_experiments": ["some_other_experiment_id"]}
    EXPECTED_STDOUT = """+-------+--------------------------------------------------------+
| ID    | some_id                                                |
+-------+--------------------------------------------------------+
| Image | tensorflow/tensorflow:latest-py3                       |
| URL   | https://aws-testing.paperspace.io/tensorboard/some_id/ |
| State | Some State                                             |
+-------+--------------------------------------------------------+
+--------------------+------------+
| Experiments ID     | State      |
+--------------------+------------+
| some_experiment_id | Some State |
+--------------------+------------+
"""

    COMMAND_WITH_API_KEY_CHANGED = [
        "tensorboards", "remove-experiments", "--id", "some_id",
        "--experiment", "some_other_experiment_id", "--apiKey", "some_key"
    ]

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"title": "Invalid credentials provided"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to update resource: Invalid credentials provided\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_send_valid_request_when_command_was_executed_with_required_options(self, put_patched):
        put_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        put_patched.assert_called_once_with(
            self.URL,
            headers=EXPECTED_HEADERS,
            json=self.EXPECTED_REQUEST_JSON,
            params=None,
            data=None,
        )
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_send_request_with_changed_api_key_when_api_key_option_was_used(self, put_patched):
        put_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=401)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_CHANGED)

        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN, result.exc_info
        put_patched.assert_called_once_with(
            self.URL,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
            json=self.EXPECTED_REQUEST_JSON,
            params=None,
            data=None,
        )


class TestTensorboardsDelete(object):
    URL = "https://services.paperspace.io/tensorboards/v1/some_id"
    COMMAND = ["tensorboards", "delete", "--id", "some_id"]

    EXPECTED_RESPONSE_JSON = example_responses.TENSORBOARD_DELETE_RESPONSE_JSON
    EXPECTED_REQUEST_JSON = {"added_experiments": [], "removed_experiments": ["some_other_experiment_id"]}
    EXPECTED_STDOUT = "Delete tensorboard (some_id) ended with success\n"

    COMMAND_WITH_API_KEY_CHANGED = [
        "tensorboards", "delete", "--id", "some_id", "--apiKey", "some_key"
    ]

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"title": "Invalid credentials provided"}
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to delete resource: Invalid credentials provided\n"

    RESPONSE_JSON_WITH_WRONG_ACCESS = {
        "error": "You don't have access to tensorboard some_id",
        "title": "401 Unauthorized"
    }
    EXPECTED_WRONG_ACCESS_RESPONSE = """Failed to delete resource: You don't have access to tensorboard some_id
401 Unauthorized
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.delete")
    def test_should_send_valid_request_when_command_was_executed_with_required_options(self, delete_patched):
        delete_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON, status_code=200)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        delete_patched.assert_called_once_with(
            self.URL,
            headers=EXPECTED_HEADERS,
            json=None,
            params=None
        )
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.delete")
    def test_should_send_request_with_changed_api_key_when_api_key_option_was_used(self, delete_patched):
        delete_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=401)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_CHANGED)

        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN, result.exc_info
        delete_patched.assert_called_once_with(
            self.URL,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
            json=None,
            params=None
        )

    @mock.patch("gradient.api_sdk.clients.http_client.requests.delete")
    def test_should_send_request_to_remove_tensorboard_without_proper_access(self, delete_patched):
        delete_patched.return_value = MockResponse(self.RESPONSE_JSON_WITH_WRONG_ACCESS, status_code=401)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_WRONG_ACCESS_RESPONSE, result.exc_info
        delete_patched.assert_called_once_with(
            self.URL,
            headers=EXPECTED_HEADERS,
            json=None,
            params=None
        )
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"



import abc
import json

import click
import six
import terminaltables
from click import style
from halo import halo

from gradient import api_sdk, exceptions, TensorboardClient
from gradient.api_sdk import constants, sdk_exceptions
from gradient.api_sdk.config import config
from gradient.api_sdk.utils import concatenate_urls
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.clilogger import CliLogger
from gradient.cliutils import none_strings_to_none_objects
from gradient.commands import tensorboards as tensorboards_commands
from gradient.commands.common import BaseCommand, ListCommandMixin, DetailsCommandMixin, StreamMetricsCommand, \
    LogsCommandMixin

try:
    # Python 3
    from itertools import zip_longest
except ImportError:
    # Python 2
    from itertools import izip_longest as zip_longest


@six.add_metaclass(abc.ABCMeta)
class BaseExperimentCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        client = api_sdk.clients.ExperimentsClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )
        return client


class TensorboardHandler(object):
    def __init__(self, api_key, logger=CliLogger()):
        self.api_key = api_key
        self.logger = logger

    def maybe_add_to_tensorboard(self, tensorboard_id, experiment_id):
        """Add experiment to existing or new tensorboard

        :param str|bool tensorboard_id:
        :param str experiment_id:
        """
        if isinstance(tensorboard_id, six.string_types):
            self._add_experiment_to_tensorboard(tensorboard_id, experiment_id)
            return

        tensorboards = self._get_tensorboards()
        if len(tensorboards) == 1:
            self._add_experiment_to_tensorboard(tensorboards[0].id, experiment_id)
        else:
            self._create_tensorboard_with_experiment(experiment_id)

    def _add_experiment_to_tensorboard(self, tensorboard_id, experiment_id):
        """Add experiment to tensorboard

        :param str tensorboard_id:
        :param str experiment_id:
        """
        command = tensorboards_commands.AddExperimentToTensorboard(api_key=self.api_key)
        command.execute(tensorboard_id, [experiment_id])

    def _get_tensorboards(self):
        """Get tensorboards

        :rtype: list[api_sdk.Tensorboard]
        """
        tensorboard_client = TensorboardClient(api_key=self.api_key, logger=self.logger)
        tensorboards = tensorboard_client.list()
        return tensorboards

    def _create_tensorboard_with_experiment(self, experiment_id):
        """Create tensorboard with experiment

        :param str experiment_id:
        """
        command = tensorboards_commands.CreateTensorboardCommand(api_key=self.api_key)
        command.execute(experiments=[experiment_id])


@six.add_metaclass(abc.ABCMeta)
class BaseCreateExperimentCommandMixin(BaseCommand):
    SPINNER_MESSAGE = "Creating new experiment"
    CREATE_SUCCESS_MESSAGE_TEMPLATE = "New experiment created with ID: {}"

    def __init__(self, workspace_handler, *args, **kwargs):
        super(BaseCreateExperimentCommandMixin, self).__init__(*args, **kwargs)
        self.workspace_handler = workspace_handler

    def execute(self, json_, add_to_tensorboard=False):
        self._handle_workspace(json_)
        self._handle_dataset_data(json_)

        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            experiment_id = self._create(json_)

        self.logger.log(self.CREATE_SUCCESS_MESSAGE_TEMPLATE.format(experiment_id))
        self.logger.log(self.get_instance_url(experiment_id, json_["project_id"]))

        self._maybe_add_to_tensorboard(add_to_tensorboard, experiment_id, self.api_key)
        return experiment_id

    def get_instance_url(self, instance_id, project_id):
        url = concatenate_urls(config.WEB_URL, "{}/projects/{}/experiments/{}".format(self.get_namespace(), project_id, instance_id))
        return url

    def _handle_workspace(self, instance_dict):
        handler = self.workspace_handler.handle(instance_dict)

        instance_dict.pop("ignore_files", None)
        instance_dict.pop("workspace", None)
        instance_dict.pop("workspace_archive", None)
        instance_dict.pop("workspace_url", None)
        if handler:
            instance_dict["workspace_url"] = handler

    def _maybe_add_to_tensorboard(self, tensorboard_id, experiment_id, api_key):
        """
        :param str|bool tensorboard_id:
        :param str experiment_id:
        :param str api_key:
        """
        if tensorboard_id is not False:
            tensorboard_handler = TensorboardHandler(api_key)
            tensorboard_handler.maybe_add_to_tensorboard(tensorboard_id, experiment_id)

    @staticmethod
    def _handle_dataset_data(json_):
        """Make list of dataset dicts"""
        datasets = [
            json_.pop("dataset_uri_list", ()),
            json_.pop("dataset_id_list", ()),
            json_.pop("dataset_name_list", ()),
            json_.pop("dataset_access_key_id_list", ()),
            json_.pop("dataset_secret_access_key_list", ()),
            json_.pop("dataset_endpoint_list", ()),
            json_.pop("dataset_version_id_list", ()),
            json_.pop("dataset_etag_list", ()),
            json_.pop("dataset_volume_kind_list", ()),
            json_.pop("dataset_volume_size_list", ()),
        ]

        if not any(datasets):
            return
        else:
            datasets_len = max(len(datasets[0]), len(datasets[1]))
            other_dataset_param_max_len = max(len(elem) for elem in datasets[2:])
            if datasets_len < other_dataset_param_max_len:
                # there no point in defining n+1 dataset parameters of one type for n datasets
                raise click.BadParameter(
                    "Too many dataset parameter sets ({}) for {} dataset URIs. Forgot to add one more dataset URI?"
                        .format(other_dataset_param_max_len, datasets_len))

        datasets = [none_strings_to_none_objects(d) for d in datasets]

        datasets = zip_longest(*datasets, fillvalue=None)
        datasets = [{"uri": dataset[0],
                     "id": dataset[1],
                     "name": dataset[2],
                     "aws_access_key_id": dataset[3],
                     "aws_secret_access_key": dataset[4],
                     "aws_endpoint": dataset[5],
                     "version_id": dataset[6],
                     "etag": dataset[7],
                     "volume_kind": dataset[8],
                     "volume_size": dataset[9],
                     } for dataset in datasets]

        json_["datasets"] = datasets

    @abc.abstractmethod
    def _create(self, json_):
        pass


class CreateSingleNodeExperimentCommand(BaseCreateExperimentCommandMixin, BaseExperimentCommand):
    def _create(self, json_):
        handle = self.client.create_single_node(**json_)
        return handle


class CreateMultiNodeExperimentCommand(BaseCreateExperimentCommandMixin, BaseExperimentCommand):
    def _create(self, json_):
        handle = self.client.create_multi_node(**json_)
        return handle


class CreateMpiMultiNodeExperimentCommand(BaseCreateExperimentCommandMixin, BaseExperimentCommand):
    def _create(self, json_):
        json_.pop("experiment_type_id", None)  # for MPI there is no experiment_type_id parameter in client method
        handle = self.client.create_mpi_multi_node(**json_)
        return handle


class CreateAndStartMultiNodeExperimentCommand(BaseCreateExperimentCommandMixin, BaseExperimentCommand):
    SPINNER_MESSAGE = "Creating and starting new experiment"
    CREATE_SUCCESS_MESSAGE_TEMPLATE = "New experiment created and started with ID: {}"

    def _create(self, json_):
        handle = self.client.run_multi_node(**json_)
        return handle


class CreateAndStartMpiMultiNodeExperimentCommand(BaseCreateExperimentCommandMixin, BaseExperimentCommand):
    SPINNER_MESSAGE = "Creating and starting new experiment"
    CREATE_SUCCESS_MESSAGE_TEMPLATE = "New experiment created and started with ID: {}"

    def _create(self, json_):
        json_.pop("experiment_type_id", None)  # for MPI there is no experiment_type_id parameter in client method
        handle = self.client.run_mpi_multi_node(**json_)
        return handle


class CreateAndStartSingleNodeExperimentCommand(BaseCreateExperimentCommandMixin, BaseExperimentCommand):
    SPINNER_MESSAGE = "Creating and starting new experiment"
    CREATE_SUCCESS_MESSAGE_TEMPLATE = "New experiment created and started with ID: {}"

    def _create(self, json_):
        handle = self.client.run_single_node(**json_)
        return handle


class StartExperimentCommand(BaseExperimentCommand):
    def execute(self, experiment_id):
        """
        :param str experiment_id:
        """
        self.client.start(experiment_id)
        self.logger.log("Experiment started")


class StopExperimentCommand(BaseExperimentCommand):
    def execute(self, experiment_id):
        """
        :param str experiment_id:
        """
        self.client.stop(experiment_id)
        self.logger.log("Experiment stopped")


class ListExperimentsCommand(ListCommandMixin, BaseExperimentCommand):
    TOTAL_ITEMS_KEY = "totalItems"

    def _get_instances(self, **kwargs):
        try:
            instances, meta_data = self.client.list(get_meta=True, **kwargs)
        except sdk_exceptions.GradientSdkError as e:
            raise exceptions.ReceivingDataFailedError(e)

        return instances, meta_data

    def _get_table_data(self, experiments):
        data = [("Name", "ID", "Status")]
        for experiment in experiments:
            name = experiment.name
            handle = experiment.id
            status = constants.ExperimentState.get_state_str(experiment.state)
            data.append((name, handle, status))
        return data

    def execute(self, **kwargs):
        return self._generate_data_table(**kwargs)


class GetExperimentCommand(DetailsCommandMixin, BaseExperimentCommand):
    def _get_table_data(self, experiment):
        """
        :param api_sdk.SingleNodeExperiment|api_sdk.MultiNodeExperiment|api_sdk.MpiMultiNodeExperiment experiment:
        """
        if experiment.experiment_type_id == constants.ExperimentType.SINGLE_NODE:
            return self._get_single_node_data(experiment)

        if experiment.experiment_type_id == constants.ExperimentType.GRPC_MULTI_NODE:
            return self._get_multi_node_grpc_data(experiment)

        if experiment.experiment_type_id == constants.ExperimentType.MPI_MULTI_NODE:
            return self._get_multi_node_mpi_data(experiment)

        raise ValueError("Wrong experiment type: {}".format(experiment.experiment_type_id))

    @staticmethod
    def _get_single_node_data(experiment):
        """
        :param api_sdk.SingleNodeExperiment experiment:
        """

        tags_string = ", ".join(experiment.tags)
        data = (
            ("Name", experiment.name),
            ("ID", experiment.id),
            ("State", constants.ExperimentState.get_state_str(experiment.state)),
            ("Ports", experiment.ports),
            ("Project ID", experiment.project_id),
            ("Worker Command", experiment.command),
            ("Worker Container", experiment.container),
            ("Worker Machine Type", experiment.machine_type),
            ("Working Directory", experiment.working_directory),
            ("Workspace URL", experiment.workspace_url),
            ("Model Type", experiment.model_type),
            ("Model Path", experiment.model_path),
            ("Tags", tags_string),
        )
        return data

    @staticmethod
    def _get_multi_node_grpc_data(experiment):
        """
        :param api_sdk.MultiNodeExperiment experiment:
        """

        tags_string = ", ".join(experiment.tags)
        data = (
            ("Name", experiment.name),
            ("ID", experiment.id),
            ("State", constants.ExperimentState.get_state_str(experiment.state)),
            ("Artifact directory", experiment.artifact_directory),
            ("Cluster ID", experiment.cluster_id),
            ("Experiment Env", experiment.experiment_env),
            ("Experiment Type", constants.ExperimentType.get_type_str(experiment.experiment_type_id)),
            ("Model Type", experiment.model_type),
            ("Model Path", experiment.model_path),
            ("Parameter Server Command", experiment.parameter_server_command),
            ("Parameter Server Container", experiment.parameter_server_container),
            ("Parameter Server Count", experiment.parameter_server_count),
            ("Parameter Server Machine Type", experiment.parameter_server_machine_type),
            ("Ports", experiment.ports),
            ("Project ID", experiment.project_id),
            ("Worker Command", experiment.worker_command),
            ("Worker Container", experiment.worker_container),
            ("Worker Count", experiment.worker_count),
            ("Worker Machine Type", experiment.worker_machine_type),
            ("Working Directory", experiment.working_directory),
            ("Workspace URL", experiment.workspace_url),
            ("Tags", tags_string),
        )
        return data

    @staticmethod
    def _get_multi_node_mpi_data(experiment):
        """
        :param api_sdk.MpiMultiNodeExperiment experiment:
        """

        tags_string = ", ".join(experiment.tags)
        data = (
            ("Name", experiment.name),
            ("ID", experiment.id),
            ("State", constants.ExperimentState.get_state_str(experiment.state)),
            ("Artifact directory", experiment.artifact_directory),
            ("Cluster ID", experiment.cluster_id),
            ("Experiment Env", experiment.experiment_env),
            ("Experiment Type", constants.ExperimentType.get_type_str(experiment.experiment_type_id)),
            ("Model Type", experiment.model_type),
            ("Model Path", experiment.model_path),
            ("Master Command", experiment.master_command),
            ("Master Container", experiment.master_container),
            ("Master Count", experiment.master_count),
            ("Master Machine Type", experiment.master_machine_type),
            ("Ports", experiment.ports),
            ("Project ID", experiment.project_id),
            ("Worker Command", experiment.worker_command),
            ("Worker Container", experiment.worker_container),
            ("Worker Count", experiment.worker_count),
            ("Worker Machine Type", experiment.worker_machine_type),
            ("Working Directory", experiment.working_directory),
            ("Workspace URL", experiment.workspace_url),
            ("Tags", tags_string),
        )
        return data


class ExperimentLogsCommand(LogsCommandMixin, BaseExperimentCommand):
    def _make_table(self, logs, experiment_id):
        table_title = "Experiment %s logs" % experiment_id
        table_data = [("JOB ID", "LINE", "MESSAGE")]
        table = terminaltables.AsciiTable(table_data, title=table_title)

        for log in logs:
            table_data.append(self._format_row(experiment_id, log))

        return table.table

    def _get_log_row_string(self, id, log):
        log_msg = "{}\t{}\t{}".format(*self._format_row(id, log))
        return log_msg

    @staticmethod
    def _format_row(experiment_id, log_row):
        return (style(fg="blue", text=experiment_id),
                style(fg="red", text=str(log_row.line)),
                log_row.message)


class DeleteExperimentCommand(BaseExperimentCommand):
    def execute(self, experiment_id, *args, **kwargs):
        self.client.delete(experiment_id)
        self.logger.log("Experiment deleted")


class ExperimentAddTagsCommand(BaseExperimentCommand):
    def execute(self, experiment_id, *args, **kwargs):
        self.client.add_tags(experiment_id, **kwargs)
        self.logger.log("Tags added to experiment")


class ExperimentRemoveTagsCommand(BaseExperimentCommand):
    def execute(self, experiment_id, *args, **kwargs):
        self.client.remove_tags(experiment_id, **kwargs)
        self.logger.log("Tags removed from experiment")


class GetExperimentMetricsCommand(BaseExperimentCommand):
    def execute(self, experiment_id, start, end, interval, built_in_metrics, *args, **kwargs):
        metrics = self.client.get_metrics(
            experiment_id,
            start=start,
            end=end,
            built_in_metrics=built_in_metrics,
            interval=interval,
        )
        formatted_metrics = json.dumps(metrics, indent=2, sort_keys=True)
        self.logger.log(formatted_metrics)

class ListExperimentMetricsCommand(BaseExperimentCommand):
    def execute(self, experiment_id, start, end, interval, *args, **kwargs):
        metrics = self.client.list_metrics(
            experiment_id,
            start=start,
            end=end,
            interval=interval,
        )
        formatted_metrics = json.dumps(metrics, indent=2, sort_keys=True)
        self.logger.log(formatted_metrics)

class StreamExperimentMetricsCommand(StreamMetricsCommand, BaseExperimentCommand):
    pass
